#!/usr/bin/python3
# -*- coding:utf-8 -*-

"""
    @file:      response.py
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @auther:    Tangmi(tangmi360@gmail.com)
    @date:      June 21, 2018
    @desc:      Response class
"""


class Response(object):

    def __init__(self):
        self._meta = {}
        self._resp = None
        self._data = None

    def __getattr__(self, attr):
        if self._resp:
            return getattr(self._resp, attr)
        else:
            object.__getattribute__(self, attr)

    @property
    def resp(self):
        return self._resp

    @resp.setter
    def resp(self, resp):
        self._resp = resp

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, data):
        self._data = data

    @property
    def meta(self):
        return self._meta

    @meta.setter
    def meta(self, meta):
        if isinstance(meta, dict):
            self._meta = meta



#!/usr/bin/python3
# -*- coding:utf-8 -*-

"""
    @file:      mysql_client.py
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @auther:    Tangmi(tangmi360@gmail.com)
    @date:      June 11, 2018
    @desc:      MySQL storage access class
"""

import logging
import asyncio
import aiomysql

from tomato.utils import singleton


@singleton
class MySQLClient(object):

    async def _create_pool(self, **kwargs):
        self._host = kwargs.get('host', 'localhost')
        self._port = kwargs.get('port', 3306)
        self._user = kwargs.get('user')
        self._passwd = kwargs.get('passwd')
        self._db = kwargs.get('db')
        self._minsize = kwargs.get('minsize', 5)
        self._maxsize = kwargs.get('maxsize', 5)
        self._charset = kwargs.get('charset', 'utf8')
        self._autocommit = kwargs.get('autocommit', True)

        loop = asyncio.get_event_loop()
        self._pool = await aiomysql.create_pool(host=self._host,
                                                port=self._port,
                                                user=self._user,
                                                password=self._passwd,
                                                db=self._db,
                                                minsize=self._minsize,
                                                maxsize=self._maxsize,
                                                charset=self._charset,
                                                autocommit=self._autocommit,
                                                loop=loop)
        async with self._pool.acquire() as conn:
            async with conn.cursor() as cur:
                await cur.execute("SELECT 1;")
                (r,) = await cur.fetchone()
                assert r == 1
                logging.info('mysql connection success: mysql://%s:%s'
                             ';charset=%s, db[%s] autocommit[%s] '
                             'minsize[%s] maxsize[%s]',
                             self._host, self._port, self._charset,
                             self._db, self._autocommit,
                             self._minsize, self._maxsize)

    def __init__(self, **kwargs):
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self._create_pool(**kwargs))

    async def close(self):
        if hasattr(self, '_pool'): # TODO 临时方案，待资源管理重构后处理
            self._pool.close()
            await self._pool.wait_closed()

    async def find(self, sql, params=None, size=0):
        """find
        """
        async with self._pool.acquire() as conn:
            async with conn.cursor(aiomysql.DictCursor) as cur:
                if params:
                    await cur.execute(sql, params)
                else:
                    await cur.execute(sql)
                if size:
                    rows = await cur.fetchmany(size)
                else:
                    rows = await cur.fetchall()
                return rows

    async def execute(self, sql, params=None, autocommit=True):
        """execute
        """
        async with self._pool.acquire() as conn:
            if not autocommit:
                await conn.begin()
            try:
                async with conn.cursor(aiomysql.DictCursor) as cur:
                    if params:
                        await cur.execute(sql, params)
                    else:
                        await cur.execute(sql)
                    affected = cur.rowcount
                if not autocommit:
                    await conn.commit()
            except BaseException as e:
                if not autocommit:
                    await conn.rollback()
                raise
            return affected



#!/usr/bin/python3
# -*- coding:utf-8 -*-

"""
    @file:      mongo_client.py
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @auther:    Tangmi(tangmi360@gmail.com)
    @date:      June 4, 2018
    @desc:      Mongodb storage access class
"""

import logging
import motor.motor_asyncio as motor

from tomato.utils import singleton


@singleton
class MongoClient(object):
    """
    Use examples

    mongodb = MongoClient(address=[10.0.0.1:27017, 10.0.0.2:27017, 10.0.0.3:27017...])
    mongodb.insert(item) // insert item
    """

    def __init__(self, address,
                 user=None,
                 passwd=None,
                 replicat_set=None,
                 min_pool_size=1,
                 max_pool_size=1,
                 auth_source=None,
                 auth_mechanism='SCRAM-SHA-1'):

        if isinstance(address, str):
            address = [address]
        self.client = motor.AsyncIOMotorClient(address, authSource=auth_source,
                                               username=user,
                                               password=passwd,
                                               replicaSet=replicat_set,
                                               minPoolSize=min_pool_size,
                                               maxPoolSize=max_pool_size,
                                               authMechanism=auth_mechanism)
        logging.info('Mongo Client init info %s', address)

    async def find_one(self, database=None,
                       collection=None,
                       key=None,
                       fields=None,
                       network_timeout=5):
        collection = self.client[database][collection]
        if fields is None:
            future = collection.find_one(key, max_time_ms=network_timeout)
        else:
            future = collection.find_one(key, fields, max_time_ms=network_timeout)
        try:
            doc = await future
        except Exception as e:
            logging.error('[%s]', str(e))
            raise
        else:
            return doc

    async def find(self, database=None,
                   collection=None,
                   key=None,
                   fields=None,
                   start=0,
                   count=0,
                   sort_item=None,
                   total=0,
                   network_timeout=5,
                   batch_size=10):
        docs = []
        collection = self.client[database][collection]
        if fields is None:
            if sort_item is not None:
                cursor = collection.find(key, max_time_ms=network_timeout, skip=start, limit=count).sort(sort_item)
            else:
                cursor = collection.find(key, max_time_ms=network_timeout, skip=start, limit=count)
        else:
            if sort_item is not None:
                cursor = collection.find(key, fields, max_time_ms=network_timeout, skip=start, limit=count).sort(sort_item)
            else:
                cursor = collection.find(key, fields, max_time_ms=network_timeout, skip=start, limit=count)
        try:
            cursor.batch_size(batch_size)
            while (await cursor.fetch_next):
                docs.append(cursor.next_object())
        except Exception as e:
            logging.error('[%s]', str(e))
            raise

        total_count = None
        if total == 1:
            cursor = collection.find(key, max_time_ms=network_timeout).count()
            try:
                total_count = await cursor
            except Exception as e:
                logging.error('[%s]', str(e))
                raise
            return (total_count, docs)
        return docs

    async def insert(self, database=None,
                     collection=None,
                     ins_item=None,
                     network_timeout=5):
        try:
            collection = self.client[database][collection]
            result = await collection.insert_one(ins_item)
        except Exception as e:
            logging.error('[%s]', str(e))
            raise
        else:
            return result

    def _data_format(self, item, result, root_prefix='', path=''):
        for key in item:
            if '$' in root_prefix or key[0] == '$':
                root_prefix = '$'
            if isinstance(item[key], dict):
                if key[0] != '$' and path != '':
                    self._data_format(item[key], result, root_prefix=root_prefix, path='{}.{}'.format(path, key))
                elif key[0] == '$' and path != '':
                    if path in result:
                        result[path][key] = {}
                    else:
                        result[path] = {key: {}}
                    self._data_format(item[key], result[path][key], root_prefix=root_prefix, path='')
                elif key[0] == '$' and path == '':
                    result[key] = {}
                    self._data_format(item[key], result[key], root_prefix=root_prefix, path='')
                else:
                    self._data_format(item[key], result, root_prefix=root_prefix, path=key)
            elif key[0] == '$' and path != '':
                result[path] = {key: item[key]}
            elif key[0] != '$' and path != '':
                if root_prefix == '$':
                    result['{}.{}'.format(path, key)] = item[key]
                else:
                    if '$set' not in result:
                        result['$set'] = {}
                    result['$set']['{}.{}'.format(path, key)] = item[key]
            else:
                if key[0] != '$' and root_prefix != '$':
                    if '$set' not in result:
                        result['$set'] = {}
                    result['$set'][key] = item[key]
                else:
                    result[key] = item[key]

    async def update(self, database=None,
                     collection=None,
                     key=None,
                     up_item=None,
                     ins_item=None,
                     upsert=False,
                     multi=False,
                     network_timeout=5):
        collection = self.client[database][collection]
        data = {}
        if up_item:
            self._data_format(up_item, data)
        if ins_item:
            data['$setOnInsert'] = ins_item
        if multi == False:
            future = collection.update_one(key, data,
                                           upsert=upsert)  # wtimeout, w
        else:
            future = collection.update_many(key, data,
                                            upsert=upsert)  # wtimeout, w
        try:
            result = await future
        except Exception as e:
            logging.error('[%s]', str(e))
            raise
        else:
            return result

    async def remove(self, database=None,
                collection=None,
                key=None,
                network_timeout=5):
        collection = self.client[database][collection]
        future = collection.remove(key) # wtimeout, w
        try:
            result = await future
        except Exception as e:
            logging.error('[%s]', str(e))
            raise
        else:
            return result



#!/usr/bin/python3
# -*- coding:utf-8 -*-

"""
    @file:      http_client.py
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @auther:    Tangmi(tangmi360@gmail.com)
    @date:      June 11, 2018
    @desc:      Basic http client functions
"""

import json
import aiohttp
import asyncio
import logging
import async_timeout

from tomato.utils import singleton
from tomato.http.request import Request
from tomato.http.response import Response


@singleton
class HttpClient(object):

    def __init__(self, max_pool_size=0, limit_per_host=0,
                 dummy_cookie=False):
        self._loop = asyncio.get_event_loop()
        self._headers = {'Accept-Encoding': 'gzip,deflate'}
        self._cookie_jar = None
        if dummy_cookie:
            self._cookie_jar = aiohttp.helper.DummyCookieJar()
        self._connector = aiohttp.TCPConnector(limit=max_pool_size,
                                               limit_per_host=limit_per_host)
        self._session = aiohttp.ClientSession(loop=self._loop, connector_owner=False,
                                              connector=self._connector,
                                              headers=self._headers,
                                              cookie_jar=self._cookie_jar)

    async def get(self, url, params=None, cookies=None,
                  headers=None, resp_type='text', timeout=10):
        # TODO cookies field not used
        async with async_timeout.timeout(timeout):
            async with self._session.get(url=url, params=params,
                                         headers=headers) as resp:
                response = await self._parse_response(resp, resp_type)
                return response

    async def post(self, url, params=None, data=None,
                   cookies=None, headers=None,
                   resp_type='text', timeout=10):
        # TODO cookies field not used
        async with async_timeout.timeout(timeout):
            async with self._session.post(url, params=params,
                                          headers=headers,
                                          data=data) as resp:
                response = await self._parse_response(resp, resp_type)
                return response

    async def _parse_response(self, resp, resp_type):
        response = Response()
        response.resp = resp
        if 'text' == resp_type:
            response.data = await resp.text()
        return response

    async def close(self):
        if not self._connector.closed:
            self._connector.close()
        await self._session.close()



#!/usr/bin/python3
# -*- coding:utf-8 -*-

"""
    @file:      request.py
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @auther:    Tangmi(tangmi360@gmail.com)
    @date:      June 21, 2018
    @desc:      Request class
"""


class Request(object):

    def __init__(self, url):
        self._url = url
        self._meta = {}

    @property
    def url(self):
        return self._url

    @property
    def meta(self):
        return self._meta

    @meta.setter
    def meta(self, meta):
        if isinstance(meta, dict):
            self._meta = meta



#!/usr/bin/python3
# -*- coding:utf-8 -*-

"""
    @file:      redis_client.py
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @auther:    Tangmi(tangmi360@gmail.com)
    @date:      June 11, 2018
    @desc:      Redis storage access class
"""

import logging
import asyncio
import aioredis

from tomato.utils import singleton


@singleton
class RedisClient(object):

    async def _create_pool(self):
        loop = asyncio.get_event_loop()
        redis_uri = 'redis://%s:%s' % (self._host, self._port)
        self._pool_dict = {}
        for db in self._db:
            pool = await aioredis.create_redis_pool(redis_uri,
                                                    db=int(db),
                                                    password=self._passwd,
                                                    minsize=self._minsize,
                                                    maxsize=self._maxsize,
                                                    loop=loop)
            self._pool_dict[int(db)] = pool
            logging.info('Redis Client Connecting to %s?db=%s', redis_uri, db)

    def __init__(self, host=None, port=None,
                   passwd=None, db='0',
                   minsize=None, maxsize=None):
        self._host = host
        self._port = port
        
        if passwd is None or len(passwd.strip()) == 0:
            self._passwd = None
        else:
            self._passwd = passwd

        self._minsize = minsize
        self._maxsize = maxsize
        self._db = db.split(',')

        loop = asyncio.get_event_loop()
        loop.run_until_complete(self._create_pool())

    async def close(self):
        if hasattr(self, '_pool_dict'): # TODO 临时方案，待资源管理重构后处理
            for db in self._pool_dict:
                self._pool_dict[db].close()
                await self._pool_dict[db].wait_closed()

    async def rpop(self, key, db=0):
        """rpop
        """
        result = await self._pool_dict[db].rpop(key)
        return result

    async def exists(self, key, db=0):
        """exists
        """
        result = await self._pool_dict[db].exists(key)
        return result

    async def set(self, key, value, db=0):
        """set
        """
        result = await self._pool_dict[db].set(key, value)
        return result

    async def get(self, key, db=0):
        """get
        """
        result = await self._pool_dict[db].get(key)
        return result

    async def lpop(self, key, db=0):
        """lpop
        """
        result = await self._pool_dict[db].lpop(key)
        return result

    async def blpop(self, key, db=0, timeout=0):
        """blpop
        """
        result = await self._pool_dict[db].blpop(key, timeout)
        return result

    async def brpop(self, key, db=0, timeout=0):
        """brpop
        """
        result = await self._pool_dict[db].brpop(key, timeout)
        return result

    async def rpush(self, key, value, db=0):
        """rpush
        """
        result = await self._pool_dict[db].rpush(key, value)
        return result

    async def lpush(self, key, value, db=0):
        """lpush
        """
        result = await self._pool_dict[db].lpush(key, value)
        return result

    async def sadd(self, key, value, db=0):
        """sadd
        """
        result = await self._pool_dict[db].sadd(key, value)
        return result

    async def smembers(self, key, db=0):
        """smembers
        """
        result = await self._pool_dict[db].smembers(key)
        return result

    async def srem(self, key, value, db=0):
        """srem
        """
        result = await self._pool_dict[db].srem(key, value)
        return result

    async def bitop(self, op, dest, key, db=0):
        """bitop
        """
        result = await self._pool_dict[db].bitop(op, dest, key)
        return result

    async def bitcount(self, key, db=0):
        """bitcount
        """
        result = await self._pool_dict[db].bitcount(key)
        return result

    async def delete(self, key, db=0):
        """delete
        """
        result = await self._pool_dict[db].delete(key)
        return result

    async def zadd(self, key, value, score, db=0):
        """zadd
        """
        result = await self._pool_dict[db].zadd(key, value, score)
        return result

    async def zrem(self, key, member, db=0):
        """zrem
        """
        result = await self._pool_dict[db].zrem(key, member)
        return result

    async def sismember(self, key, value, db=0):
        """sismember
        """
        result = await self._pool_dict[db].sismember(key, value)
        return result

    async def zrangebyscore(self, key, min, max, db=0, withscores=False):
        """zrangebyscore
        """
        result = await self._pool_dict[db].zrangebyscore(name=key, min=min,
                                                         max=max, withscores=withscores)
        return result

    async def zincrby(self, key, member, incrment, db=0):
        """zincrby
        """
        result = await self._pool_dict[db].zincrby(key, member, incrment)
        return result

    async def hset(self, key, field, value, db=0):
        """hset
        """
        result = await self._pool_dict[db].hset(key, field, value)
        return result

    async def hsetnx(self, key, field, value, db=0):
        """hsetnx
        """
        result = await self._pool_dict[db].hsetnx(key, field, value)
        return result

    async def hget(self, key, field, db=0):
        """hget
        """
        result = await self._pool_dict[db].hget(key, field)
        return result
    
    async def hmget(self, key, field, db=0):
        """hmget
        """
        result = await self._pool_dict[db].hmget(key, field)
        return result

    async def hdel(self, key, field, db=0):
        """hdel
        """
        result = await self._pool_dict[db].hdel(key, field)
        return result
        
    async def llen(self, key, db=0):
        """llen
        """
        result = await self._pool_dict[db].llen(key)
        return result

    async def hlen(self, key, db=0):
        """hlen
        """
        result = await self._pool_dict[db].hlen(key)
        return result

    async def scard(self, key, db=0):
        """scard
        """
        result = await self._pool_dict[db].scard(key)
        return result

    async def zcard(self, key, db=0):
        """zcard
        """
        result = await self._pool_dict[db].zcard(key)
        return result
    
    async def lrange(self, key, start, end, db=0):
        """lrange
        """
        result = await self._pool_dict[db].lrange(key, start, end)
        return result



#!/usr/bin/python3
# -*- coding:utf-8 -*-

"""
    @file:      http_server.py
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @auther:    Tangmi(tangmi360@gmail.com)
    @date:      June 11, 2018
    @desc:      Basic http server functions and router rules encapsulation
"""

import json
import signal
import logging
import asyncio
import functools
from aiohttp import web

from tomato.utils import util
from tomato.utils import singleton


@singleton
class HttpServer(object):

    def __init__(self, host, port):
        self._loop = asyncio.get_event_loop()
        self._host = host
        self._port = int(port)
        self._handlers = {'/': self.default_handler}

    async def default_handler(self, request):
        req_params = request.query
        response = 'Welcome to tomato'
        return response

    def init_handler(self, conf_file=None):
        logging.info('init handlers...')
        conf = util.load_conf(conf_file)
        secs = conf.sections()
        for sec in secs:
            handler = conf.get(sec, 'name')
            router = conf.get(sec, 'router')
            module_name = '.'.join(handler.split('.')[0:-1])
            class_name = handler.split('.')[-1]
            cls_obj = util.gen_object(module_name=module_name, class_name=class_name)
            for attr in dir(cls_obj):
                # 加载非__前缀的属性
                if attr[0] != '_':
                    # 获取导入obj方法。
                    class_attr_obj = getattr(cls_obj, attr)
                    # 判断类属性是否为函数
                    if hasattr(class_attr_obj, '__call__'):
                        # TODO 未判断函数是否有装饰器
                        # TODO 尾段如果是{}未做处理
                        # TODO (\w+)的正则匹配判断
                        abs_router = router
                        handler_name = router.split('/')[-1]
                        if '(' in router and ')' in router:
                            begin_idx = router.index('(')
                            end_idx = router.index(')')
                            replace_str = router[begin_idx:end_idx+1]
                            abs_router = router.replace(replace_str, attr)
                            handler_name = attr

                        if attr == 'setup':
                            logging.info('execute call[%s:%s]', handler, attr)
                            self._loop.run_until_complete(class_attr_obj())
                        elif attr == handler_name:
                            if abs_router not in self._handlers:
                                self._handlers[abs_router] = class_attr_obj
                                logging.info('init router info: router[%s--->%s:%s]',
                                             abs_router, handler, handler_name)
                            else:
                                logging.warning('duplicate router configuration: '
                                                'router[%s--->%s:%s]',
                                                abs_router, handler, handler_name)

        logging.info('handler initialization is complete')

    @web.middleware
    async def data_codec(self, request, handler):
        """http简单解包与封包
           1.qs信息从request.query属性中直接获取（MultiDictProxy类型）
           2.http body内容如果请求头为content_type=application/json,
             则通过json标准库进行解析，并覆盖request.body，
             否则直接把body内容抛到业务层自行处理（默认是byte类型）
           3.对于response的处理，目前简单判断为dict则直接转成json格式str发给调用端
             否则直接按业务层返回值返回给调用端
        """
        request.body = None
        if request.body_exists and request.can_read_body:
            request.body = await request.content.read()
            if request.content_type == 'application/json':
                try:
                    request.body = json.loads(request.body)
                except Exception as e:
                    logging.warning('json format warning, errmsg[%s]', str(e))
                    return web.json_response({'ret': 0, 'msg': 'params error'})
        response = await handler(request)
        if isinstance(response, dict):
            return web.json_response(response)
        else:
            return web.Response(body=response)

    async def init_server(self):
        self._app = web.Application(middlewares=[self.data_codec,])

        for router in self._handlers:
            self._app.router.add_route('*', router, self._handlers[router])

        self._runner = web.AppRunner(self._app)
        await self._runner.setup()
        site = web.TCPSite(self._runner, self._host, self._port)
        await site.start()
        logging.info('serving on [%s]', site.name)

    async def start(self):
        await self.init_server()

    async def close(self):
        await self._runner.cleanup()
        await self._app.shutdown()
        await self._app.cleanup()
